#include <cstdio>
#include <algorithm>
#include <map>
#include <assert.h>

#include "resolution.h"
#include "groupscreen.h"
#include "fetch_group.h"
#include "fetch_max_score_for_songs.h"
#include "fetch_max_score_for_players.h"
#include "fetch_needs_update.h"
#include "fetch_highscore.h"
#include "fonts.h"

GroupScreen::GroupScreen(pqxx::connection &conn, unsigned tournament, unsigned round, unsigned parallel, unsigned machine, unsigned num_machines, unsigned players_per_machine)
	: tournament(tournament), round(round), parallel(parallel), machine(machine), num_machines(num_machines), players_per_machine(players_per_machine), scores_changed(conn, "scores"), conn(conn), valid(false)
{
}

GroupScreen::~GroupScreen()
{
}

bool GroupScreen::check_invalidated()
{
	if (!valid)
		return true;
	if (!scores_changed.get_flag())
		return false;

	bool needs_update;
	conn.perform(FetchNeedsUpdate(last_updated, tournament, round, parallel, &needs_update));

	if (!needs_update)
		scores_changed.reset_flag();
	
	return needs_update;
}

unsigned GroupScreen::get_show_players(const Group &group)
{
	unsigned num_players_this_machine = (group.players.size() + num_machines - machine - 1) / num_machines;
	return std::min(num_players_this_machine, 9U);
}

void GroupScreen::draw_main_heading(std::vector<TextDefer> &td)
{
	char heading[64];
	if (num_machines == 1) {
		if (parallel == 0) {
			std::sprintf(heading, "Round %u", round);
		} else {
			std::sprintf(heading, "Round %u, Group %u", round, parallel);
		}
	} else {
		if (parallel == 0) {
			std::sprintf(heading, "Round %u, Machine %u", round, machine + 1);
		} else {
			std::sprintf(heading, "Round %u, Group %u, Machine %u", round, parallel, machine + 1);
		}
	}

	unsigned width = my_draw_text(heading, NULL, 40.0);
	my_draw_text_deferred(td, heading, 40.0, LOGICAL_SCREEN_WIDTH/2 - width/2, 60);
}

// make column headings from the first player's songs
void GroupScreen::draw_column_headings(std::vector<TextDefer> &td, const Group &group, const std::vector<unsigned> &colwidth)
{
	unsigned num_scores = group.players[0].scores.size();

	unsigned col = 1;
	unsigned x = 40 + colwidth[0];
	for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i, ++col) {
		if (!i->chosen) {
			unsigned this_width = my_draw_text(i->song.short_title, NULL, 12.0);
			my_draw_text_deferred(td, i->song.short_title, 12.0, x + colwidth[col] / 2 - this_width / 2, 100);
		}
		x += colwidth[col] + 20;
	}

	if (num_scores > 1) {
		my_draw_text_deferred(td, "Total", 12.0, x + colwidth[num_scores + 1] / 2 - my_draw_text("Total", NULL, 12.0) / 2, 100);
		x += colwidth[num_scores + 1] + 20;
	}
	my_draw_text_deferred(td, "Rank", 12.0, x + colwidth[num_scores + 2] / 2 - my_draw_text("Rank", NULL, 12.0) / 2, 100);
}	
	
// show all the players and the scores
void GroupScreen::draw_scores(std::vector<TextDefer> &td, const Group &group, unsigned min_player, const std::vector<unsigned> &colwidth)
{
	unsigned max_num_width = my_draw_text("8888", NULL, 22.0);
	unsigned num_scores = group.players[0].scores.size();
	unsigned show_players = get_show_players(group);
	unsigned y = (show_players <= 7) ? 140 : (140 - (show_players - 7) * 5);
	
	unsigned row = 0, m = 0, x;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end() && row < 9; ++i) {
		if (m++ % num_machines != machine)
			continue;
		if (m-1 < min_player)
			continue;

		my_draw_text_deferred(td, i->nick, 18.0, 20, y);

		x = 40 + colwidth[0];

		unsigned col = 1;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			char text[16] = "";
			if (j->score != -1) {
				std::sprintf(text, "%u", j->score);
			}
	
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			if (j->chosen) {
				my_draw_text_deferred(td, text, 22.0, x + max_num_width - this_width, y);

				// draw the long name if we can, otherwise use the short one
				if (my_draw_text(j->song.title, NULL, 12.0) > (colwidth[col] - 10 - max_num_width)) {
					my_draw_text_deferred(td, j->song.short_title, 12.0, x + max_num_width + 10, y);
				} else {
					my_draw_text_deferred(td, j->song.title, 12.0, x + max_num_width + 10, y);
				}
			} else {
				my_draw_text_deferred(td, text, 22.0, x + colwidth[col] / 2 - this_width / 2, y);
			}
			x += colwidth[col] + 20;
		}

		// draw total
		if (num_scores > 1) {
			char text[16];
			std::sprintf(text, "%u", i->total);
			
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			my_draw_text_deferred(td, text, 22.0, x + colwidth[num_scores + 1] / 2 - this_width / 2, y);
			x += colwidth[num_scores + 1] + 20;
		}

		if (show_players > 7)
			y += 40 - (show_players - 7) * 4;
		else 
			y += 40;
		++row;
	}
}	

/*
 * Find out how wide each column has to be. First try unlimited width (ie.
 * long titles for everything); if that gets too long, try again with short
 * titles for chosen songs.
 */
void GroupScreen::find_column_widths(const Group &group, std::vector<unsigned> &colwidth)
{
	unsigned num_scores;
	unsigned max_num_width = my_draw_text("8888", NULL, 22.0);
	unsigned sumcolwidth;
	
	for (unsigned mode = 0; mode < 2; ++mode) {
		for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
			unsigned col = 1;
			
			if (colwidth.size() == 0)
				colwidth.push_back(0);
			
			colwidth[0] = std::max(colwidth[0], my_draw_text(i->nick, NULL, 18.0));

			for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
				if (colwidth.size() < col+1)
					colwidth.push_back(0);
					
				if (j->chosen) {
					colwidth[col] = std::max(colwidth[col], my_draw_text((mode == 0) ? j->song.title : j->song.short_title, NULL, 12.0) + 
							max_num_width + 10);
				} else {		
					colwidth[col] = std::max(colwidth[col], my_draw_text(j->song.short_title, NULL, 12.0));
					colwidth[col] = std::max(colwidth[col], max_num_width);
				}
			}
		}

		num_scores = group.players[0].scores.size();

		if (colwidth.size() < num_scores + 2) {
			colwidth.push_back(0);
			colwidth.push_back(0);
		}
	
		if (num_scores > 1) {
			colwidth[num_scores + 1] = std::max(my_draw_text("Total", NULL, 12.0), max_num_width);
		}
		colwidth[num_scores + 2] = my_draw_text("Rank", NULL, 12.0);

		// if we're at long titles and that works, don't try the short ones
		sumcolwidth = 0;
			
		for (unsigned i = 0; i <= num_scores + 2; ++i)
			sumcolwidth += colwidth[i] + 20;
		
		if (sumcolwidth < 780)
			break;

		if (mode == 0) {
			colwidth.erase(colwidth.begin(), colwidth.end());
		}
	}

	/* 
	 * If we have space to go, distribute as much as we can to the chosen song column, so we won't have
	 * total and rank jumping around.
	 */
	if (sumcolwidth < 780) {
		int first_chosen_col = -1;
		unsigned col = 1;

		for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i, ++col) {
			if (i->chosen) {
				first_chosen_col = col;
				break;
			}
		}

		if (first_chosen_col != -1) {
			colwidth[first_chosen_col] += 780 - sumcolwidth;
		}
	}
}

/* Find the first player with the fewest songs played and part of this machine. */
const Player *GroupScreen::get_next_player(const Group &group)
{
	unsigned min_played_songs = 9999;
	const Player *next_player = NULL;
	unsigned m = 0;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		unsigned this_played = 0;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j) {
			if (j->score != -1)
				++this_played;
		}

		if ((m++ % num_machines == machine) && this_played < min_played_songs) {
			min_played_songs = this_played;
			next_player = &(*i);
		}
	}

	return next_player;
}

/*
 * At the bottom, for a single player, is "who's playing, what will he/she be
 * playing, and optionally, how much to lead/win and how much to secure
 * qualification" (the last one only in the final round). We assume playing is
 * done in a modified zigzag; all the random songs are played first in
 * zigzag/wrapping order (player 1 song 1, player 2 song 2, player 3 song 3,
 * player 1 song 2, player 2 song 3, player 3 song 1, etc... assuming three
 * songs and three players) and then all the chosen songs are played (we assume
 * only one chosen song).
 *
 * The lines are as follows:
 *
 * <player>
 * <song>
 * High score: <hs> by <hsplayer> at <hsevent>
 * Needs to lead: <leadscore>
 * Needs to secure qualification: <qualscore>
 * Needs to win group: <winscore>
 */
void GroupScreen::draw_next_up_single(unsigned char *buf, const Group &group,
	std::map<unsigned, unsigned> &song_scores, std::map<unsigned, unsigned> &player_scores,
	const std::vector<unsigned> &max_score, const std::vector<unsigned> &min_score)
{
        unsigned num_scores = group.players[0].scores.size();
	
	// Find out how many random songs there are (equal for all players).
	unsigned num_random_songs = 0;
	for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i) {
		if (!i->chosen)
			++num_random_songs;
	}

	/* 
	 * Find out which player is next, and what song he she is supposed to play. First
	 * try random songs.
	 */
	const Player *next_player = get_next_player(group);
	const Score *next_song = NULL;

	for (unsigned i = 0; i < num_random_songs; ++i) {
		unsigned j = (i + next_player->position - 1) % num_random_songs;
		if (next_player->scores[j].score == -1) {
			next_song = &(next_player->scores[j]);
			break;
		}
	}

	// then all songs, if that didn't work out (slightly icky, but hey)
	if (next_song == NULL) {
		for (unsigned i = 0; i < num_scores; ++i) {
			unsigned j = (i + next_player->position) % num_scores;
			if (next_player->scores[j].score == -1) {
				next_song = &(next_player->scores[j]);
				break;
			}
		}
	}

	if (next_song != NULL) {
		// find out how many songs we've played in all
		unsigned num_played = 0;
		for (unsigned i = 0; i < num_scores; ++i) {
			if (next_player->scores[i].score != -1) {
				++num_played;
			}
		}
	
		bool last_song = (num_played == num_scores - 1);
			
		draw_next_up_player(buf, group, *next_player, *next_song, last_song, song_scores, player_scores, max_score, min_score);
	}
}

/*
 * Some tournaments allow versus play in the initial rounds to save time; this is
 * of course for random songs only. In this case, the scheme from draw_next_up_single()
 * is somewhat changed, as we zig-zag across pairs instead of players. (If there's a
 * stray person left in the group, that player plays the song by him-/herself as in
 * a usual single tournament.
 */
void GroupScreen::draw_next_up_versus(unsigned char *buf, const Group &group,
	std::map<unsigned, unsigned> &song_scores, std::map<unsigned, unsigned> &player_scores,
	const std::vector<unsigned> &max_score, const std::vector<unsigned> &min_score)
{
	// Find out how many random songs there are (equal for all players).
	unsigned num_random_songs = 0;
	for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i) {
		if (!i->chosen)
			++num_random_songs;
	}

	// Find the next player and what song he/she is supposed to play, if any.
	const Player *next_player = get_next_player(group);
	const Score *next_song = NULL;
	unsigned song_num;

	for (unsigned i = 0; i < num_random_songs; ++i) {
		unsigned j = (i + (next_player->position - 1) / 2) % num_random_songs;
		if (next_player->scores[j].score == -1) {
			next_song = &(next_player->scores[j]);
			song_num = j;
			break;
		}
	}
	
	/*
	 * If there's no match, we're on the chosen songs (or done),
	 * so just delegate to draw_up_single().
	 */
	if (next_song == NULL) {
		draw_next_up_single(buf, group, song_scores, player_scores, max_score, min_score);
		return;
	}
	
	/*
	 * Look for a player with the same amount of random songs played _and_ missing
	 * the same song.
	 */ 
	unsigned num_songs_played = 0;
	for (unsigned i = 0; i < num_random_songs; ++i) {
		if (next_player->scores[i].score != -1) {
			++num_songs_played;
		}
	}
	
	unsigned m = 0;
	const Player *other_player = NULL;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		if ((m++ % num_machines != machine))
			continue;
		if (i->id == next_player->id)
			continue;
		
		unsigned this_songs_played = 0;
		for (unsigned j = 0; j < num_random_songs; ++j) {
			if (i->scores[j].score != -1) {
				++this_songs_played;
			}
		}

		if (this_songs_played != num_songs_played)
			continue;

		if (i->scores[song_num].score == -1) {
			other_player = &(*i);
			break;
		}	
	}

	// If we didn't find another player, just draw the one we have as usual.
	if (other_player == NULL) {
		draw_next_up_player(buf, group, *next_player, *next_song, false,
			song_scores, player_scores, max_score, min_score);
		return;
	}
	
	// OK, we have two players. Draw their nicks and the scores
	widestring text = widestring("Next players: ") + next_player->nick + widestring(" and ") + other_player->nick;
	unsigned this_width = my_draw_text(text, NULL, 24.0);
	my_draw_text(text, buf, 24.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 420);

	if (next_song->song.id != -1) {
		this_width = my_draw_text(next_song->song.title, NULL, 20.0);
		my_draw_text(next_song->song.title, buf, 20.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 457);

		Highscore hs;
		conn.perform(FetchHighscore(next_song->song.id, &hs));

		if (hs.score != -1) {
			text = widestring("High score: ") + widestring(pqxx::to_string(hs.score)) +
				widestring(", by ") + hs.nick + widestring(" in ") + hs.tournament_name;
			this_width = my_draw_text(text, NULL, 16.0);
			my_draw_text(text, buf, 16.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 487);
		}
	}
}

void GroupScreen::draw_next_up_player(unsigned char *buf, const Group &group, const Player &player, const Score &song, bool last_song,
	std::map<unsigned, unsigned> &song_scores, std::map<unsigned, unsigned> &player_scores,
	const std::vector<unsigned> &max_score, const std::vector<unsigned> &min_score)
{
	widestring text = widestring("Next player: ") + player.nick;
	unsigned this_width = my_draw_text(text, NULL, 24.0);
	my_draw_text(text, buf, 24.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 420);

	if (song.song.id != -1) {
		this_width = my_draw_text(song.song.title, NULL, 20.0);
		my_draw_text(song.song.title, buf, 20.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 457);

		Highscore hs;
		conn.perform(FetchHighscore(song.song.id, &hs));

		if (hs.score != -1) {
			text = widestring("High score: ") + widestring(pqxx::to_string(hs.score)) +
				widestring(", by ") + hs.nick + widestring(" in ") + hs.tournament_name;
			this_width = my_draw_text(text, NULL, 16.0);
			my_draw_text(text, buf, 16.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, 487);
		}
	}

	// only show lead/win/qualify for the last song
	if (last_song) {
		/*
		 * Find out how much we need to lead, how much we need to be guaranteed
		 * to win the group, and how much we need to secure qualification.
		 */

		// find the best score we can get
		unsigned max_score_this_song;
		if (song.song.id != -1) {
			// random song, or we know what song the player picked
			max_score_this_song = song_scores[song.song.id];
		} else {
			max_score_this_song = player_scores[player.id];
		}

		unsigned y = 520;

		// see what score this player must beat to lead
		unsigned lead_beat = 0, win_beat = 0;
		for (unsigned i = 0; i < group.players.size(); ++i) {
			if (group.players[i].id == player.id)
				continue;

			lead_beat = std::max(lead_beat, group.players[i].total);
		}

		// find the best max score among the others
		for (unsigned i = 0; i < group.players.size(); ++i) {
			if (group.players[i].id == player.id)
				continue;

			win_beat = std::max(win_beat, max_score[i]);
		}

		/*
		 * There's a somewhat subtle point here. Normally, what a player would be interested in
		 * with regard to qualification would be a set of three values:
		 *
		 * 1. How much is the absolute minimum required to qualify, given that all others
		 *    fail?
		 * 2. How much will give a reasonable chance of qualifying, given the expected performance
		 *    of all the others?
		 * 3. How much will be enough to secure qualification, no matter what?
		 *
		 * Given perfect guessing, #2 would be "how much is needed to qualify"; however, it is
		 * completely impossible to give an exact value for that, and we're not into the guessing
		 * games. :-) #1 is often so low it's completely unrealistic (ie. far enough from #2 that
		 * it's not interesting), but #3, the most conservative estimate, is often a good measure.
		 * #3 is "how much is needed to _secure_ qualification", and that is usually what we
		 * print out when it's possible.
		 *
		 * However, in a few situations, #1 and #3 will be the exact same value, from which it
		 * follows (from the squeeze law, or just common sense :-) ) that #2 will be the same
		 * value as #1 and #3. (This usually happens near or at the end of a group.) In that
		 * case, we know the value we seek (ie. "how much is needed to qualify"), so we drop
		 * the word "secure" and just print it as-is.
		 *
		 * To find #1 and #3, we sort and pick out the values we need to beat in the best and
		 * the worst case.
		 */
		int qualify_beat_worst_case = -1, qualify_beat_best_case = -1;

		if (group.num_qualifying > 0) {
			std::vector<unsigned> tmp;

			for (unsigned i = 0; i < group.players.size(); ++i) {
				if (group.players[i].id == player.id)
					continue;
				tmp.push_back(max_score[i]);
			}
			std::sort(tmp.begin(), tmp.end());
			qualify_beat_worst_case = tmp[tmp.size() - group.num_qualifying];

			std::vector<unsigned> tmp2;
			for (unsigned i = 0; i < group.players.size(); ++i) {
				if (group.players[i].id == player.id)
					continue;
				tmp2.push_back(min_score[i]);
			}

			std::sort(tmp2.begin(), tmp2.end());
			qualify_beat_best_case = tmp2[tmp2.size() - group.num_qualifying];
		}

		// print out the lines we can attain
		if (player.total + max_score_this_song > lead_beat && (lead_beat != win_beat)) {
			int lead_need = std::max(lead_beat - player.total + 1, 0U);

			if (lead_need > 1) {
				text = widestring("Needs to lead: ") + widestring(pqxx::to_string(lead_need));
				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, y);

				y += 30;
			}
		}

		if (player.total + max_score_this_song > win_beat) {
			int win_need = std::max(win_beat - player.total + 1, 0U);

			if (win_need > 0) {
				text = widestring("Needs to win: ") + widestring(pqxx::to_string(win_need));

				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, y);

				y += 30;
			}
		}

		if (group.num_qualifying > 0 &&
		    group.num_qualifying != group.players.size() &&
				player.total + max_score_this_song > unsigned(qualify_beat_worst_case) &&
				(unsigned(qualify_beat_worst_case) != win_beat)) {
			int qual_need = std::max(qualify_beat_worst_case - player.total + 1, 0U);

			if (qual_need > 0) {
				if (qualify_beat_worst_case == qualify_beat_best_case) {
					text = widestring("Needs to qualify: ") + widestring(pqxx::to_string(qual_need));
				} else {
					text = widestring("Needs to secure qualification: ") + widestring(pqxx::to_string(qual_need));
				}

				this_width = my_draw_text(text, NULL, 18.0);
				my_draw_text(text, buf, 18.0, (LOGICAL_SCREEN_WIDTH/2) - this_width/2, y);

				y += 30;
			}
		}
	}
}

// some refactoring done, more should be
void GroupScreen::draw(unsigned char *buf, unsigned width, unsigned height)
{
	std::vector<TextDefer> td;
	
	scores_changed.reset_flag();
	set_screen_size(width, height);

	/*
	 * We'll probably need some values from here later on (although not all), just fetch them
	 * all while we're at it.
	 */
	std::map<unsigned, unsigned> song_scores, player_scores;
	conn.perform(FetchMaxScoreForSongs(tournament, &song_scores));
	conn.perform(FetchMaxScoreForPlayers(tournament, round, &player_scores));
	
	Group group;
	conn.perform(FetchGroup(tournament, round, parallel, &group));
	gettimeofday(&last_updated, NULL);

	memset(buf, 0, width * height * 4);

	std::vector<unsigned> colwidth;
	
	draw_main_heading(td);
	find_column_widths(group, colwidth);
	draw_column_headings(td, group, colwidth);

	// Find out which player is next. we want to show SHOW_PLAYERS players, centered
	// around this as much as possible. (Usually, this will mean all, but not always.)
	unsigned show_players = get_show_players(group);
	const Player *center_player = get_next_player(group);
	
	// find the index (kind of backwards...)
	int player_index = -1;
	for (unsigned i = 0; i < group.players.size(); ++i) {
		if (&(group.players[i]) == center_player) {
			player_index = i;
			break;
		}
	}

	assert(player_index >= 0);

	int min_player = player_index - signed(show_players) / 2;
	if (min_player + show_players > group.players.size()) // FIXME: songs_per_machine
		min_player = group.players.size() - show_players;
	if (min_player < 0)
		min_player = 0;

	draw_scores(td, group, min_player, colwidth);
	
	unsigned num_scores = group.players[0].scores.size();

	/*
	 * Approximate (but probably working quite well in practice) heuristic
	 * for finding the min and max rank of a player works as follows:
	 *
	 * First of all, find out, for each player in the group, what the
	 * maximum remaining score possibly can be (the minimum score is of
	 * course identical to the player's current total). For a random song,
	 * this is of course 1000 * (maximum feet rating) (but of course, that
	 * depends on whether we can play single or double! for now, assume
	 * double is okay, but this logic will be deferred to FetchMaxScore
	 * anyhow); for a random song, we simply pick the highest-ranking song
	 * we can find, EXCEPT those the player has chosen earlier AND the
	 * random songs this round, AND all random songs from elimination rounds
	 * (ie. rounds with only one group). (Phew!) This doesn't solve problems
	 * we'd face with more than one chosen song, but it should be good enough.
	 *
	 * After we've found the max and min scores for all players, it's a simple
	 * matter of sorting; the best attainable rank for player X is obtained if 
	 * X gets max score and all others get min score, the worst attainable rank
	 * is obtained if X gets min score and all others get max score.
	 */
	std::vector<unsigned> max_score, min_score;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		unsigned min_score_tp = 0, max_score_tp = 0;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j) {
			if (j->score != -1) {
				// already given
				min_score_tp += j->score;
				max_score_tp += j->score;
			} else {
				unsigned max_score_this_song;
				if (j->song.id != -1) {
					// random song, or we know what song the player picked
					max_score_this_song = song_scores[j->song.id];
				} else {
					max_score_this_song = player_scores[i->id];
				}
				max_score_tp += max_score_this_song;
			}
		}
		max_score.push_back(max_score_tp);
		min_score.push_back(min_score_tp);
	}
	
	// now finally find min and max rank, and draw it all
	unsigned y = (show_players <= 7) ? 140 : (140 - (show_players - 7) * 5);
	for (unsigned i = 0; i < group.players.size() && (i/num_machines) < show_players+min_player; ++i) {
		unsigned best_rank = 1, worst_rank = 1;
		for (unsigned j = 0; j < group.players.size(); ++j) {
			if (i == j)
				continue;

			if (max_score[i] < min_score[j])
				++best_rank;
			if (min_score[i] <= max_score[j])
				++worst_rank;
		}

		char text[16];
		if (best_rank == worst_rank)
			std::sprintf(text, "%u", best_rank);
		else
			std::sprintf(text, "%u-%u", best_rank, worst_rank);
		
		if (i % num_machines != machine)
			continue;
		if (signed(i) < min_player)
			continue;
		
		// find out where to place this
		unsigned x = 40 + colwidth[0];
		for (unsigned j = 1; j <= num_scores + 1; ++j)
			x += colwidth[j] + 20;

		// minor correction :-)
		if (num_scores <= 1)
			x -= 20;
		
		unsigned this_width = my_draw_text(text, NULL, 22.0);
		my_draw_text_deferred(td, text, 22.0, x + colwidth[num_scores + 2] / 2 - this_width / 2, y);

		if (show_players > 7)
			y += 40 - (show_players - 7) * 4;
		else 
			y += 40;
	}

	if (players_per_machine == 2)
		draw_next_up_versus(buf, group, song_scores, player_scores, max_score, min_score);
	else 
		draw_next_up_single(buf, group, song_scores, player_scores, max_score, min_score);
	
	valid = true;
	draw_all_deferred_text(buf, td, last_text);
	last_text = td;
}

int GroupScreen::get_priority()
{
	return 10;
}
